//*****************************************************************************
//*
//*		
//*		BandPass 
//*
//*
//*****************************************************************************

#include	<math.h>
#include	"BandPass.h"

#define 	CHANNEL_MAX		2
#define 	EQULIZER_MAX	16
#define		PI				3.14159265358979

static	double	C_coef[EQULIZER_MAX*CHANNEL_MAX];
static	double	C_ener[EQULIZER_MAX*CHANNEL_MAX];
static	double	Volume[EQULIZER_MAX*CHANNEL_MAX+CHANNEL_MAX];

static	int		eq_count	= 0;

//*****************************************************************************
//*
//*		BandPassInit
//*
//*****************************************************************************
//	Frequenzen der einzelnen Bnder angeben
//	Freq		: Zeiger auf ein Feld mit den Frequenzen
//	Count		: Anzahl der Frequenzen im Feld
//	Fs			: Ist die Samplefrequens der Eingansdaten an BandPass(...)
void BandPassInit(const double *Freq,int Count,double Fs)
{
int		x,y;
double	dt,c;


	if(Count>=EQULIZER_MAX)Count=EQULIZER_MAX-1;

	eq_count = Count;
	dt = 1.0 / Fs;

	for(x=0;x<Count;x++)
		{
		c = 1.0 / (2.0*PI*Freq[x]);
		C_coef[x] = dt / c;

		for(y=1;y<CHANNEL_MAX;y++)
			{
			C_coef[x+y*EQULIZER_MAX] = C_coef[x];
			}

	}

	

}


//*****************************************************************************
//*
//*		BandPassVolume
//*
//*****************************************************************************
//	Setzt die Lautstrke der einzelnen Frequenzen
//	Vol			: Zeiger auf ein Feld mit den Lausstrken.Das Feld muss die 
//				  bergebene Frequenzanzahl plus 1 aus BandPassInit(...) haben.
//				  Der letzte Eintrag ist die Lautstrke.
//				  Es sollten Werte zwischen 0.0 und 3.0 benutzt werden.
//	Channel		: Fr welchen Kanal sind die Einstellungen. (0=links 1=rechts)
void BandPassVolume(const double *Vol,int Channel)
{
int x,y;


	for(x=0,y=Channel*EQULIZER_MAX;x<=eq_count;x++,y++)
		{
		Volume[y] = Vol[x];
		}

}


//*****************************************************************************
//*
//*		BandPass
//*
//*****************************************************************************
//	Filtert eingelsesene Sampels.
//	Output		: Speicher fr die Ausgabe.
//	Input		: Zeiger auf die Eingangsdaten
//	Count		: Anzahl der Sampels in den Eingangsdaten (pro Kanal)
//	Channels	: Anzahl der Kanle in den Eingangsdaten
void BandPass(short *Output,const short *Input,int Count,int Channels)
{
int			x,y,ch,eq;
double		uin,duc;
double		u[EQULIZER_MAX];
	
	
	
	for(x=0;x<Count;)
		{
		for(ch=0;ch<Channels;ch++,x++)
			{
			uin = Input[x];

	
			eq = ch*EQULIZER_MAX;
					
			for(y=0;y<eq_count;y++,eq++)	// Einzel Bnder berechnen
				{
				duc	=  (uin-C_ener[eq])*C_coef[eq];	
				
				u[y]  = C_ener[eq];
				uin  -= C_ener[eq];

				C_ener[eq] += duc;
				}

			uin *= Volume[eq];
			eq = ch*EQULIZER_MAX;

			for(y=0;y<eq_count;y++,eq++)	// Bnder addieren
				{
				uin += u[y]*Volume[eq++];
				}

			if(uin> 32767.0)uin= 32767.0;
			if(uin<-32768.0)uin=-32768.0;

			Output[x] = (short)uin; 
			}
		}



}