//*****************************************************************************
//*
//*
//*		WavClass
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2007
//
#include	<string.h>
#include	"WavClass.h"


//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
WavClass::WavClass()
{
	Init();
}

//*****************************************************************************
//*
//*		Destruktor
//*
//*****************************************************************************
WavClass::~WavClass()
{
	Close();
}

//*****************************************************************************
//*
//*		Open
//*
//*****************************************************************************
//	ffnet eine WAV Datei und liest den Header ein
//	Ist Exact!=0 mssen alle Eintge im Header stimmen
//	Ergibt	: 0 wenn alles ok bzw. 1 bei einem Fehler
int  WavClass::Open(const char *FileName,int Exact)
{
int			 x;
FILE		*file;

	if(mode)Close();


		file=fopen(FileName,"rb");
	if(!file)return 3;

	ifile=file;
	free_ifile=1;
	mode=1;


		x=ReadHeader();						// Header laden
	if(!x || (!Exact && x==2))return 0;

	Close();


return 1;
}

//*****************************************************************************
//*
//*		Create
//*
//*****************************************************************************
//	ffnet eine WAV Datei und liest den Header ein
//	Ergibt	0 wenn ok sontst 1
int  WavClass::Create(const char *FileName)
{
FILE	*file;


	if(mode)Close();


		file=fopen(FileName,"wb");
	if(!file)return 1;

	ofile=file;
	free_ofile=1;
	mode=2;

return 0;
}
//*****************************************************************************
//*
//*		Close
//*
//*****************************************************************************
//	Schliet WAV Datei und liest den Header enien
//	Ergibt	0 wenn ok sontst 1
int  WavClass::Close()
{
int		x,ok=0;
char	buffer[4];
long	len;


	if(mode==0)return 1;


	if(mode==1)
		{
		for(x=0;x<chunk_count;x++)					// Chunks lschen
			delete chunk_ptr[x];
		chunk_count=0;

		if(free_ifile && ifile)						// Mit Open erzeugt ?
			{
			ok=(fclose(ifile)!=0);
			}
		Init();
		return ok;
		}

	if(mode==2)
		{
		if(write_header==0)WriteHeader();

		if(write_header!=2)
			{
			if(free_ofile)							// Mit Open erzeugt ?
				{
				ok=(fclose(ofile)!=0);
				delete ofile;
				}
			Init();
			return 1;
			}

		x=1;
		fseek(ofile,sample_offset+sample_count*byte_per_sample+pos_leak,SEEK_SET);

		if(pos_leak)								// Sample aufrunden
			{
			x=1+byte_per_sample-pos_leak;
			sample_pos++;
			if(sample_pos>sample_count)sample_count=sample_pos;
			}

		if(fwrite("\00\00\00\00\00\00\00\00\00\00\00\00\00\00\00\00",x,1,ofile)!=1)ok=1;

		len=sample_count*byte_per_sample;			// 'data' Lnge
		buffer[0] = (unsigned char)(len);
		buffer[1] = (unsigned char)(len>>8);
		buffer[2] = (unsigned char)(len>>16);
		buffer[3] = (unsigned char)(len>>24);

		if(fseek(ofile,sample_offset-4,SEEK_SET))ok=1;
		else{
			if(fwrite(buffer,4,1,ofile)!=1)ok=1;
			}

		len+=sample_offset+1-8;						// 'RIFF' Lnge
		buffer[0] = (unsigned char)(len);
		buffer[1] = (unsigned char)(len>>8);
		buffer[2] = (unsigned char)(len>>16);
		buffer[3] = (unsigned char)(len>>24);

		if(fseek(ofile,4,SEEK_SET))ok=1;
		else{
			if(fwrite(buffer,4,1,ofile)!=1)ok=1;
			}

		for(x=0;x<chunk_count;x++)					// Schreibe Chunks
			{
			if(!chunk_new[x])continue;
			if(fseek(ofile,chunk_pos[x],SEEK_SET))ok=1;
			else{
				if(fwrite(chunk_ptr[x],chunk_len[x],1,ofile))ok=1;
				}
			}

		for(x=0;x<chunk_count;x++)					// Chunks lschen
			delete chunk_ptr[x];
		chunk_count=0;


		if(free_ofile)								// Mit Create erzeugt ?
			{
			if(fclose(ofile))ok=1;
			ofile=0;
			}

		Init();
		return ok;
		}


return 1;
}

//*****************************************************************************
//*
//*		Init
//*
//*****************************************************************************
void WavClass::Init()
{
	mode			= 0;				// 1=Lesen 2=Schreiben 0=Leer
	sample_offset	= 0;				// Offset vom Sample Beginn
	sample_pos		= 0;				// Aktuelle Schreib/Leseposition
	sample_count	= 0;				// Anzahl der gespeicherten Samples
	sample_len		= 0;				// Anzahl der gespeicherten Byte im data Bereich
	sample_bits		= 16;				// Anzahl der Bits
	sample_rate		= 44100;			// Samplerate in Hz
	sample_stream	= 44100*4;			// Samplerate in Hz
	sample_blocklen = 4;				// Blocklnge
	sample_channels	= 2;				// Anzahl der Kanle
	byte_per_sample	= 4;				// Bytes je Sample
	free_ifile		= 0;				// Ausgabe Stream freigeben
	free_ofile		= 0;				// Ausgabe Stream freigeben
	chunk_count		= 0;				// Anzahl der Chunks
	pos_leak		= 0;				// Restliche Bytes vom Sample
	pos_write		=-1;				// Letzte Schreibposition
	write_header	= 0;				// Ist Header gespeichert
	null_cut        = 0;				// Keine Null-Unterdrckung
	ifile           = 0;
	ofile           = 0;
}


//*****************************************************************************
//*
//*		ReadHeader
//*
//*****************************************************************************
//	Liest den WAV-Header ein
//	Ergibt	: 0 wenn alles ok bzw.
//			  1 bei einem falschen Header
//			  2 bei falsche ChunkLnge
//			  3 bei Lese-Fehler
int WavClass::ReadHeader()
{
unsigned char  *buffer;
unsigned char	chunk[1024];
unsigned long 	len,end,pos,anz;
int				x,y,fmt_load=0;
int				fmt_num=-1;


	fseek(ifile,0,SEEK_SET);
	pos=0;

	for(x=0;x<CWC_CHUNK_MAX;x++)					// Lade Chunks
		{
		y = fread(chunk,1,8,ifile);

		if(y!=8)
			{
			if(y==0 && pos==end)break;
			return 3;
			}

		len=(chunk[4])|(chunk[5]<<8)|(chunk[6]<<16)|(chunk[7]<<24);
		pos+=8;

		if(x==0)									// Suche RIFF-Chunk
			{
			if(memcmp(chunk,"RIFF",4)    )return 1;
			if(fread (chunk,4,1,ifile)!=1)return 3;
			if(memcmp(chunk,"WAVE",4)    )return 1;
			end=len+8;
			pos+=4;
			continue;
			}

		if(chunk[0]<' ')return 2;					// Prfe Schreibweise
		if(chunk[1]<' ')return 2;
		if(chunk[2]<' ')return 2;
		if(chunk[3]<' ')return 2;
		if(chunk[0]>'z')return 2;
		if(chunk[1]>'z')return 2;
		if(chunk[2]>'z')return 2;
		if(chunk[3]>'z')return 2;


		if(!memcmp(chunk,"data",4))					// Sample Anfang
			{
			sample_len   =len;
			sample_offset=pos;

			if(pos+len+1==end)break;
			if(pos+len+1>=end)return 2;
			
			if(!fseek(ifile,len,SEEK_CUR))
				{
				pos += len;
				continue;
				}

			while(len>0)							// Falls Seek nicht geht
				{
				anz=(len>=sizeof(chunk))? sizeof(chunk):len;
				if(fread(chunk,anz,1,ifile)!=1)return 3;
				len-=anz;
				pos+=anz;
				}

			continue;
			}

		if(!memcmp(chunk,"fmt ",4) && len<CWC_CHUNK_LEN)// fmt-Chunk
			{
			fmt_load=1;
			fmt_num =chunk_count;
			}

		if(chunk_count-fmt_load>=CWC_CHUNK_MAX-1 || // Zuviele Chunks ?
			len>CWC_CHUNK_LEN)						// Chunk zu lang ?
			{
				pos+=len;
			if(!fseek(ifile,len,SEEK_CUR))continue;

			while(len>0)							// Falls Seek nicht geht
				{
				anz=(len>=sizeof(chunk))? sizeof(chunk):len;
				if(fread(chunk,anz,1,ifile)!=1)return 3;
				len-=anz;
				pos+=anz;
				}

			continue;
			}

		buffer=new unsigned char[len+1];			// Chunck Daten lesen
		buffer[len]=0;

		if(fread(buffer,len,1,ifile)!=1){delete buffer;return 3;}

		memcpy(chunk_name+(chunk_count<<2),chunk,4);
		chunk_pos[chunk_count]=pos;
		chunk_len[chunk_count]=len;
		chunk_ptr[chunk_count]=(char*)buffer;
		chunk_count++;

		pos+=len;
		}


	if(fmt_num==-1)return 1;						// fmt-Chunk gefunden ?


	buffer=(unsigned char*)chunk_ptr[fmt_num];

	sample_channels	= buffer[ 2]+(buffer[ 3]<<8);
	sample_rate		= buffer[ 4]+(buffer[ 5]<<8)+(buffer[ 6]<<16)+(buffer[ 7]<<24);
	sample_stream	= buffer[ 8]+(buffer[ 9]<<8)+(buffer[10]<<16)+(buffer[11]<<24);
	sample_blocklen = buffer[12]+(buffer[13]<<8);
	sample_bits		= buffer[14]+(buffer[15]<<8);
	byte_per_sample	= sample_blocklen;
	sample_count	=(end-pos)/(unsigned long)byte_per_sample;


return 0;
}

//*****************************************************************************
//*
//*		Attach
//*
//*****************************************************************************
//	Fgt ein File der Klasse hinzu
//	Ergibt	0 wenn ok sontst 1
int WavClass::Attach(FILE *File,int WriteMode,int Exact)
{
int x;


	if(mode!=0)Close();

	if(WriteMode==0)
		{
		ifile=File;
		mode=1;

			x=ReadHeader();						// Header laden
		if(!x || (!Exact && x==2))return 0;

		Close();
		return 1;
		}


return 1;
}

//*****************************************************************************
//*
//*		Read
//*
//*****************************************************************************
//	Ld die Anzahl Count an Byte in Buffer
//	Ergibt die Anzahl der geladen Bytes bzw. -1 bei einem Fehler
int WavClass::Read(void *Buffer,int Count)
{
int	x,y;



	if(mode!=1)return -1;						// Falscher Modus



	   x = fread((char*)Buffer,1,Count,ifile);
	if(x<=0)return -1;


	y =(unsigned long)x+pos_leak;				// Vorschub
	pos_leak    = y%byte_per_sample;
	sample_pos += y/byte_per_sample;


return x;
}
//*****************************************************************************
//*
//*		Read
//*
//*****************************************************************************
//	Ld die Anzahl Count an Byte und speicher sie in IFile
//	Ergibt die Anzahl der geladen Bytes bzw. -1 bei einem Fehler
int WavClass::Read(FILE *OFile,int Count)
{
int		anz,len,y;
char	buffer[CWC_STREAM_BUFFER];



	if(mode!=2)return -1;						// Falscher Modus

	anz=0;

	while(Count>0)
		{
		len=(Count>=sizeof(buffer))? sizeof(buffer):Count;

		   y=Read(buffer,len);					// Aus Wav lesen
		if(y==-1)
			{
			if(anz==0)anz=-1;
			break;
			}

		if(fwrite(buffer,y,1,OFile)!=1)			// Aus Stream schreiben
			{
			if(anz==0)anz=-1;
			break;
			}

		Count-=y;
		anz	 +=y;
		}


return anz;
}

//*****************************************************************************
//*
//*		Seek
//*
//*****************************************************************************
//	Verschiebt den Schreib/Lesezeiger auf eine neue Sample Position
//	Mode	: 0		Bezogen auf Anfangs-Position
//			: 1		Bezogen auf jetzige Position
//			: 2		Bezogen auf End-Position
//	Ergibt die neue Position bzw -1 bei einem Fehler
long WavClass::Seek(long Sample,int Mode)
{


	if(Mode==1)Sample+=sample_pos;				// Modus
	if(Mode==2)Sample+=sample_count;

	if(Sample<0)Sample=0;

	if(mode==1)									// Lese -Modus
		{
		if(Sample>(long)sample_count)Sample=sample_count;
		if(fseek(ifile,sample_offset+sample_count*byte_per_sample,SEEK_SET))return -1;

		sample_pos=Sample;

		return Sample;
		}

	if(mode==2)									// Schreib -Modus
		{
		fflush(ofile);
		if(fseek(ofile,sample_offset+sample_count*byte_per_sample,SEEK_SET))return -1;

		pos_write=Sample;

		return Sample;
		}

return -1;
}

//*****************************************************************************
//*
//*		SetMode
//*
//*****************************************************************************
//	Modus einstellen
//	Ergibt 0 wenn alles ok bzw 1 bei einem Fehler
int WavClass::SetMode(unsigned long SampleRate,
					  unsigned int  BytsPerSampe,
				 	  unsigned int  Stereo,
				 	  unsigned int  BlockLen)
{

	if(mode!=2     )return 1;
	if(write_header)return 1;

	if(SampleRate>50000)return 1;
	if(BytsPerSampe>16 )return 1;
	if(BytsPerSampe<1  )return 1;
	if(BlockLen==0     )BlockLen=BytsPerSampe;

	byte_per_sample = BytsPerSampe;				// Bytes je Sample
	sample_rate		= SampleRate;				// Samplerate in Hz
	sample_stream	= SampleRate*BytsPerSampe;	// Samplerate fr Stream
	sample_blocklen	= BlockLen;					// Blocklnge
	sample_channels =(Stereo)? 2:1;				// Anzahl der Kanle
	sample_bits		= byte_per_sample*8/sample_channels;


return 0;
}


//*****************************************************************************
//*
//*		WriteHeader
//*
//*****************************************************************************
//	Schreibt einen WAV-Header
//	Ergibt	: 0 wenn alles ok bzw. 1 bei einem Fehler
int WavClass::WriteHeader()
{
int				x;
unsigned char   buffer[16];
unsigned long 	pos,len;



	if(write_header)return 1;						// Schon gespeichert
	   write_header=1;


	   fseek(ofile,0,SEEK_SET);
	if(fwrite("RIFF\00\00\00\x01WAVEfmt \x10\00\00\x00",20,1,ofile)!=1)return 1;


	buffer[ 0] = (unsigned char)(1);
	buffer[ 1] = (unsigned char)(0);
	buffer[ 2] = (unsigned char)(sample_channels);
	buffer[ 3] = (unsigned char)(0);
	buffer[ 4] = (unsigned char)(sample_rate);
	buffer[ 5] = (unsigned char)(sample_rate>>8);
	buffer[ 6] = (unsigned char)(sample_rate>>16);
	buffer[ 7] = (unsigned char)(sample_rate>>24);
	buffer[ 8] = (unsigned char)(sample_stream);
	buffer[ 9] = (unsigned char)(sample_stream>>8);
	buffer[10] = (unsigned char)(sample_stream>>16);
	buffer[11] = (unsigned char)(sample_stream>>24);
	buffer[12] = (unsigned char)(sample_blocklen);
	buffer[13] = (unsigned char)(sample_blocklen>>8);
	buffer[14] = (unsigned char)(byte_per_sample*8/sample_channels);
	buffer[15] = (unsigned char)(0);

	if(fwrite(buffer,16,1,ofile)!=1)return 1;

	pos=20+16;

	for(x=0;x<chunk_count;x++)						// Schreibe Chunks
		{
		if(!memcmp(chunk_name+(x<<2),"fmt ",4))continue;
		if(!memcmp(chunk_name+(x<<2),"data",4))continue;

		len=chunk_len[x];
		memcpy(buffer,chunk_name+(x<<2),4);
		buffer[4] = (unsigned char)(len);
		buffer[5] = (unsigned char)(len>>8);
		buffer[6] = (unsigned char)(len>>16);
		buffer[7] = (unsigned char)(len>>24);

		if(fwrite(buffer,8,1,ofile)!=1)return 1;
		pos+=8;

		chunk_pos[x]=pos;
		chunk_new[x]=0;

		if(fwrite(chunk_ptr[x],len,1,ofile)!=1)return 1;
		pos+=len;
		}

	if(fwrite("data\00\00\00\x01",8,1,ofile)!=1)return 1;

	sample_offset=pos+8;
	write_header=2;


return 0;
}
//*****************************************************************************
//*
//*		Write
//*
//*****************************************************************************
//	Schreibt die Anzahl Count an Byte aus Buffer
//	Ergibt die Anzahl der gespeicherten Bytes bzw. -1 bei einem Fehler
int WavClass::Write(void *Buffer,int Count)
{
int		y,add=0;



	if(mode!=2)return -1;						// Falscher Modus

	if(write_header!=2)							// Header speichern
		{
		y=Count;
		if(null_cut)NullCut(Buffer,Count);
		if(Count<=0)return y;
		if(WriteHeader())return -1;
		add=y-Count;
		}

	if(pos_write!=-1)							// Gab es einen Verschub
		{
		sample_pos=pos_write;
		pos_leak=0;
		pos_write=-1;
		}

	if(fwrite(Buffer,Count,1,ofile)!=1)return -1;

	sample_len+=Count;

	y =(unsigned long)Count+pos_leak;			// Vorschub
	pos_leak    = y%byte_per_sample;
	sample_pos += y/byte_per_sample;


	if(sample_count<sample_pos)sample_count=sample_pos;


return Count+add;
}

//*****************************************************************************
//*
//*		Write
//*
//*****************************************************************************
//	Schreibt die Anzahl Count an Bytes aus IFile
//	Ergibt die Anzahl der geladen Bytes bzw. -1 bei einem Fehler
int WavClass::Write(FILE *IFile,int Count)
{
int		anz,len,y;
char	buffer[CWC_STREAM_BUFFER];



	if(mode!=2)return -1;						// Falscher Modus

	anz=0;

	while(Count>0)
		{
		len=(Count>=sizeof(buffer))? sizeof(buffer):Count;

		len=fread(buffer,1,len,IFile);			// Aus Stream lesen
		if(len<=0)
			{
			if(anz==0)anz=-1;
			break;
			}

		y=Write(buffer,len);

		if(y!=len)	{							// Fehler
					if(y !=-1)anz+=y;
					if(anz==0)anz=-1;
					break;
					}

		Count     -=len;
		anz	      +=len;
		sample_len+=len;
		}


return anz;
}

//*****************************************************************************
//*
//*		ChunkName
//*
//*****************************************************************************
//	Speichert den Chunk-Namen Nr im 5-Byte Buffer Name
//	Ergibte 0 wenn Name kopiert wurde sonst 1
int	WavClass::ChunkName(char *Name,int Nr)
{

	if(Nr>=chunk_count || Nr<0)
		{
		memset(Name,0,5);
		return 1;
		}
	memcpy(Name,chunk_name+(Nr<<2),4);
	Name[4]=0;

return 0;
}

//*****************************************************************************
//*
//*		ChunkFind
//*
//*****************************************************************************
//	Sucht einen Chunk-Namen
//	Ergibt -1 wenn Name nicht gefunden wurde sonst die Nummer
int	WavClass::ChunkFind(const char *Name)
{
int x;


	for(x=chunk_count-1;x>=0;x--)
		if(!memcmp(Name,chunk_name+(x<<2),4))break;

return x;
}

//*****************************************************************************
//*
//*		GetChunk
//*
//*****************************************************************************
//	Kopiert den Chunk Nr nach Mem, Len ist die Lnge Von Len
//	Ergibte 1 wenn Chunk nicht gefunden wurde sonst 0
int	WavClass::GetChunk(int Nr,void *Mem,int Len)
{
int max;


	if(Nr>=chunk_count || Nr<0)
		{
		if(Len>0)*(char*)Mem=0;
		return 1;
		}

	max=(chunk_len[Nr]<Len)? chunk_len[Nr]:Len;
	if(max< 0 ) max=0;
	if(max<Len)((char*)Mem)[max]=0;
	memcpy(Mem,chunk_ptr[Nr],max);

return 0;
}

//*****************************************************************************
//*
//*		GetChunkPos
//*
//*****************************************************************************
//	Ergibt die Position des Chunk nach dem Namen bzw. Lngenangaben in der Datei
int	WavClass::GetChunkPos(int Nr)
{

	if(Nr>=chunk_count || Nr<0)return -1;

return chunk_pos[Nr];
}

int	WavClass::GetChunkPos(const char *Name)
{
	if(!strcmp(Name,"data"))return sample_offset;

return GetChunkPos(ChunkFind(Name));
}

//*****************************************************************************
//*
//*		GetChunkLen
//*
//*****************************************************************************
//	Ergibt die Lnge des Chunk ohne dem Namen bzw. Lngenangaben in der Datei
int	WavClass::GetChunkLen(int Nr)
{

	if(Nr>=chunk_count || Nr<0)return -1;

return chunk_len[Nr];
}

int	WavClass::GetChunkLen(const char *Name)
{
	if(!strcmp(Name,"data"))return sample_len;

return GetChunkPos(ChunkFind(Name));
}

//*****************************************************************************
//*
//*		AddChunk
//*
//*****************************************************************************
//	Kopiert den Chunk Nr nach Mem, Len ist die Lnge Von Len
//	Ergibte 1 wenn nicht angefgt wurde sonst 0
int	WavClass::AddChunk(const char *Name,void *Mem,int Len)
{
int		x,nr;
void   *alt;


	if(mode!=2)return -1;						// Falscher Modus

	Len=(Len+3)&~3;

	for(x=0;x<4;x++)							// Prfe Namen
		{
		if(Name[x]<' ')return 1;
		if(Name[x]>'z')return 1;
		}

	if(*(long*)Name==*(long*)"RIFF")return 1;	// Unerlaubter Chunk
	if(*(long*)Name==*(long*)"data")return 1;
	if(*(long*)Name==*(long*)"fmt ")return 1;


	nr=ChunkFind(Name);							// Namen suchen
	alt=0;

	if(nr<0)									// Neuer Chunk
		{
		if(write_header!=0)return 1;			// Keine Neuen Chunks erlaubt

		if(chunk_count >= CWC_CHUNK_MAX)return 1;
		nr=chunk_count;
		   chunk_count++;

		memcpy(chunk_name+(nr<<2),Name,4);		// Neuer Name
		chunk_len[nr]=Len;
		chunk_ptr[nr]=new char[Len+1];
		chunk_pos[nr]=-1;						// Chunk Position
		chunk_new[nr]= 0;						// Chunk neu speichern
		}
	else{
		if(chunk_pos[nr]>0)chunk_new[nr]=1;		// Chunk neu speichern
		}


	memcpy(chunk_ptr[nr],Mem,Len);				// Daten kopieren
	chunk_ptr[nr][Len]=0;

return 0;
}



//*****************************************************************************
//*
//*		NullCut
//*
//*****************************************************************************
//	Zum einstellen der Null-Unterdrckung am Start einstellen
//	Ist Amplitude == 0 wird die Null-Unterdrckung abgeschalten
void WavClass::NullCut(int Amplitude)
{
	if(Amplitude<0)
			null_cut=-Amplitude;
	else	null_cut= Amplitude;
}
int WavClass::NullCut(lpvoid &Buffer,int &Count)
{
int		 max,pos,amp_p,amp_m;
short	*ptr2;
char	*ptr1;


	if(Count<=0)return 0;						// Leerer Buffer

	amp_p=      null_cut;
	amp_m=-(int)null_cut;

	switch(byte_per_sample/sample_channels)
			{
	case 2:	ptr2=(short*)Buffer;
			max=Count>>1;
			for(pos=0;pos<max;pos++)
				{
				if(ptr2[pos]>=amp_p)break;
				if(ptr2[pos]<=amp_m)break;
				}

			pos/=sample_channels;
			pos*=sample_channels;

			if(pos<2048)pos=0;
			Count-=pos +pos;
			Buffer=ptr2+pos;

			if(pos!=max)null_cut=0;

			return pos+pos;

	case 1:	ptr1=(char*)Buffer;
			max=Count>>1;
			amp_p>>=1;
			amp_m>>=1;

			for(pos=0;pos<max;pos++)
				{
				if(ptr1[pos]>=amp_p)break;
				if(ptr1[pos]<=amp_m)break;
				}


			pos/=sample_channels;
			pos*=sample_channels;

			if(pos<2048)pos=0;
			Count-=pos;
			Buffer=ptr1+pos;

			if(pos!=max)null_cut=0;

			return pos;

	default:null_cut=0;							// Falsches Format
			return 0;
			}

return 0;
}
