//*****************************************************************************
//*
//*
//*		WavClass	und		WaveConvert Stream
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2007
//

#include	<stdio.h>
#include	<stdlib.h>


#define 	CWC_CHUNK_MAX		8		// Maximale Anzahl an Chunks
#define 	CWC_CHUNK_LEN	   (4096*4)	// Maximale Gre eines Chunks
#define 	CWC_STREAM_BUFFER   4096    // Gre des Stream-Puffers


//*****************************************************************************
//*
//*		Klasse
//*
//*****************************************************************************
class	WavClass
{
public:
	void NullCut(int Hig);

	 WavClass();
	~WavClass();

	int  Create(const char *FileName);		// Datei erzeugen zum lesen
	int  Open  (const char *F,int Exact=0);	// Datei ffnen
	int  Read  (void *Buffer ,int Count);	// Bytes lesen
	int  Read  (FILE *OFile  ,int Count);	// Bytes lesen
	int  Write (void *Buffer ,int Count);	// Bytes speichren
	int  Write (FILE *IFile  ,int Count);	// Bytes speichren
	int  Close ();							// Datei schlieen

	int	 ChunkName(      char *Name,int Nr);// Chunk-Namen holen
	int	 ChunkFind(const char *Name);		// Chunk-Namen suchen
	int	 GetChunk (const char *Name,void *Mem,int Len);	// Chunk holen
	int  AddChunk (const char *Name,void *Mem,int Len);	// Chunk speichern
	int	 GetChunk (int   Nr  ,void *Mem,int Len);		// Chunk holen
	int	 GetChunkPos(const char *Name);		// Chunk Position holen
	int	 GetChunkPos(int Nr);				// Chunk Position holen
	int	 GetChunkLen(const char *Name);		// Chunk Lnge holen
	int	 GetChunkLen(int Nr);				// Chunk Lnge holen

	int  Attach(FILE *File,int WriteMode=0,int Exact=0);
	int  Detach();							// Stream abdocken

	long Seek(long Sample,int Mode=0);		// Lese/Schreibzeiger ndern
	long Tell();							// Lese/Schreibzeiger lesen

	int	 SetMode(unsigned long SampleRate=44100,// Modus einstellen
				 unsigned int  BytsPerSampe=4,
				 unsigned int  Stereo=1,
				 unsigned int  BlockLen=0);

	unsigned long	Rate();				// Sampelfrequenz abfragen
	unsigned long	Count();			// Anzahl der Sampels abfragen (ist nicht die Byte Lnge)
	unsigned long	StreamRate();		// StreamRate abfragen (Byte je Sekunde) = Frequenz*ByteJeSample
	unsigned int	BytePerSample();	// Bytes je Sample abfragen = Bits*Channels/8
	unsigned int	BlockLength();		// Ist 0 fr unkomprimierte Sampels
	unsigned int	Channels();			// Anzahl der Kanle abfragen
	unsigned int	Bits();				// Bits fr einen Wert abfragen

protected:

typedef	void *lpvoid;

	void Init();						// Variabeln initialisieren
	int  ReadHeader();					// Header einlesen
	int  WriteHeader();					// Header schreiben
	int  NullCut(lpvoid &Buffer,int &Count);

//*****************************************************************************
//*
//*		Variabeln
//*
//*****************************************************************************
protected:


	int				mode;				// 1=Lesen 2=Schreiben 0=Leer
	unsigned long	sample_offset;		// Offset vom Sample Beginn
	unsigned long	sample_pos;			// Aktuelle Schreib/Leseposition
	unsigned long	sample_bits;		// Bitanzahl fr ein Sample auf einem Kanal
	unsigned long	sample_count;		// Anzahl der gespeicherten Samples
	unsigned long	sample_rate;		// Samplerate in Hz (Samples/s)
	unsigned long	sample_stream;		// Samplerate fr Stream (Bytes/s)
	unsigned long	sample_len;			// Anzahl der gespeicherten Byte im data Bereich
	unsigned int	sample_channels;	// Anzahl der Kanle
	unsigned int	sample_blocklen;	// Blocklnge (Byte/Sample)
	unsigned int	byte_per_sample;	// Bytes je Sample
	unsigned int	null_cut;			// Null-Unterdrckung

	bool			free_ifile;			// Ausgabe File freigeben
	bool			free_ofile;			// Ausgabe File freigeben

	unsigned int	write_header;		// Ist Header gespeichert
	unsigned int	pos_leak;			// Restliche Bytes
			 long 	pos_write;			// Letzte Schreibposition


	FILE		   *ifile;				// Einlese File
	FILE		   *ofile;				// Ausgabe File

	int		chunk_count;				// Anzahl der Chunks
	int		chunk_len [CWC_CHUNK_MAX  ];// Chunk Lnge
	int		chunk_pos [CWC_CHUNK_MAX  ];// Chunk Position
	int		chunk_new [CWC_CHUNK_MAX  ];// Chunk neu speichern
	char	chunk_name[CWC_CHUNK_MAX*4];// Chunk Namen
	char   *chunk_ptr [CWC_CHUNK_MAX  ];// Chunk Speicher

};


//*****************************************************************************
//*
//*		Inline Funktionen
//*
//*****************************************************************************
inline int				WavClass::Detach()		 {return Close();}
inline long				WavClass::Tell()		 {return sample_pos;}
inline unsigned long	WavClass::Rate()		 {return sample_rate;}
inline unsigned long	WavClass::Count()		 {return sample_count;}
inline unsigned long	WavClass::StreamRate()	 {return sample_stream;}
inline unsigned int 	WavClass::Bits()		 {return sample_bits;}
inline unsigned int		WavClass::Channels()	 {return sample_channels;}
inline unsigned int		WavClass::BlockLength()	 {return sample_blocklen;}
inline unsigned int		WavClass::BytePerSample(){return byte_per_sample;}
inline       	int		WavClass::GetChunk   (const char *Name,void *Mem,int Len){return GetChunk(ChunkFind(Name),Mem,Len);}












