/*****************************************************************************
/*
/*		A audio driver for windows
/*
/*		Writte by Anton Zechen (C)2000
/*
/*****************************************************************************/
/*
/*	(C) Copyright Anton Zechner 2007
/*/
#include	<io.h>
#include	<stdio.h>
#include	<windows.h>
#include	<windowsx.h>
#include	<mmsystem.h>
#include	"WinRecord.h"

/*	Example:

	FILE *f=fopen("rec.dat","wb");
	RecordOpen(44100,16,2,0x4000);
	RecordScanInputs();
	RecordSetInputEnable(REC_MIC);
	RecordSetInputVolume(REC_MIC,32000);
	RecordReadFileThread(f);
	Sleep(10000);
	fclose(f);
	RecordStop();
	RecordClose();

*/

#ifndef	WIN32
#error	"This Audio-Record-Driver is only for Windows !"
#endif

#pragma	comment( lib, "winmm.lib" )


#define 	REC_BUFFER_MIN		16			/* Min buffer count      */
#define		REC_BUFFER_MAX		32			/* Max buffer count      */
#define 	REC_MIXERLINE_MAX	16			/* Max line buffer count */

static  CRITICAL_SECTION  rec_cs={0};
static  WAVEFORMATEX	  rec_wm;
static  WAVEINCAPS		  rec_device;
static  HWAVEIN			  rec_id    = 0;
static  HANDLE			  rec_event = 0;
static  HANDLE			  rec_thread= 0;
static  HMIXER			  rec_mixer = 0;
static  UINT			  rec_close = 0;
static  UINT			  rec_mixer_id_en;
static  UINT			  rec_mixer_id_vol;
static	INT				  rec_line_count = -1;
static	MIXERLINE		  rec_line[REC_MIXERLINE_MAX];
static  WAVEHDR          *rec_htr [REC_BUFFER_MAX];
static	char 		     *rec_ptr [REC_BUFFER_MAX];
static  unsigned int	  rec_count = 0;		/* buffer count     */
static  unsigned int	  rec_th_ex = 0;		/* thread exit      */
static  unsigned int	  rec_write = 0;		/* write buffer nr  */
static  unsigned int	  rec_read  = 0;		/* read buffer nr   */
static  unsigned int	  rec_wait  = 0;		/* wait for input   */
static  unsigned int	  rec_len	= 0;		/* bytes write      */
static  unsigned int	  rec_on    = 0;		/* is recording on  */
static  unsigned int	  rec_proc  = 0;		/* run write proc   */
static           int	  rec_handle=-1;		/* file handle      */
static  unsigned int      rec_bsize = 0x4000;	/* Buffer size      */


/*****************************************************************************
/*
/*		RecordInProc
/*
/*****************************************************************************
/*	Callback function for reording */
static void CALLBACK RecordInProc(	HWAVEOUT hwo,
									UINT uMsg,
									DWORD dwInstance,
									DWORD dwParam1,
									DWORD dwParam2 )
{
unsigned next;
int		 delta;
WAVEHDR	*hdr;
MMRESULT result;


switch(uMsg)
		{
case WIM_DATA:

		hdr=(WAVEHDR*)dwParam1;

		EnterCriticalSection(&rec_cs);

		   rec_read=hdr->dwUser;
		   next=rec_read+1;
		if(next>=rec_count)next=0;

		if(rec_close==0)
			{
			   result = waveInAddBuffer( rec_id, rec_htr[next], sizeof(WAVEHDR) );
			if(result== MMSYSERR_NOERROR)rec_read+=next;
			}

		   delta= rec_read - rec_write;
		if(delta< 0)delta += rec_count;
		if(delta>=1 && rec_wait)SetEvent(rec_event);

		LeaveCriticalSection(&rec_cs);

		break;

/*case WIM_CLOSE:	break;*/
/*case WIM_OPEN:	break;*/
		}

}

/*****************************************************************************
/*
/*		RecordProcThread
/*
/*****************************************************************************/
static	int (_cdecl *rrpt_proc)(void*,void*,int);
static DWORD WINAPI RecordProcThread(LPVOID *Param)
{
int exit;
rec_th_ex = 0;
exit = RecordReadProc(rrpt_proc,Param);
rec_th_ex = 1;
return exit;
}

/*****************************************************************************
/*
/*		RecordFileWrite
/*
/*****************************************************************************/
static int _cdecl RecordFileWrite(void *File,void *Buffer,int Size)
{
return fwrite(Buffer,1,Size,(FILE*)File);
}

/*****************************************************************************
/*
/*		RecordOpen
/*
/*****************************************************************************
/*	Opens a new audio input
/*	Freq		: Is the sampling frequency (8000,11025,22050,44100)
/*	Bits		: Is the bit-count for one value (8,12,16)
/*	Channels	: Is the number of output-channels (1=mono,2=stereo)
/*	BufferSize	: Is the maximum buffer size for one write operation
/*	Return	0	: All is ok.
/*			1	: No devices
/*			2	: Memmory error
/*			3	: Open error
/*			4	: Global error
/*			5	: Devices error			*/
int RecordOpen(	unsigned  Freq,
				unsigned  Bits,
				unsigned  Channels,
				unsigned  BufferSize )
{
MMRESULT	result;
unsigned	x;



if(rec_id)RecordClose();


rec_write=0;
rec_read =0;
rec_wait =0;
rec_on   =0;

if(BufferSize)rec_bsize=BufferSize;


   result = waveInGetNumDevs();
if(result == 0)return 1;
   result =waveInGetDevCaps (0, &rec_device, sizeof(WAVEINCAPS));
if(result!=MMSYSERR_NOERROR)return 5;


rec_wm.cbSize          = 0;
rec_wm.wFormatTag      = WAVE_FORMAT_PCM;
rec_wm.nChannels       = Channels;
rec_wm.wBitsPerSample  = Bits;
rec_wm.nSamplesPerSec  = Freq;
rec_wm.nAvgBytesPerSec = Freq * Channels * Bits/8;
rec_wm.nBlockAlign     =        Channels * Bits/8;



   result = waveInOpen( &rec_id,0, &rec_wm,(DWORD)RecordInProc ,0 ,CALLBACK_FUNCTION );
/*if(result== MMSYSERR_ALLOCATED)
	{
	waveInClose(rec_id);
	result = waveInOpen( &rec_id,0, &rec_wm,(DWORD)RecordInProc ,0 ,CALLBACK_FUNCTION );
	}
*/
if(result!= MMSYSERR_NOERROR)return 3;


rec_event  = CreateEvent(NULL,FALSE,FALSE,0);
InitializeCriticalSection(&rec_cs);

for(rec_count=0; rec_count < REC_BUFFER_MAX; rec_count++)
	{
	    rec_htr[rec_count] = (LPWAVEHDR)GlobalAllocPtr(GMEM_MOVEABLE | GMEM_SHARE,sizeof(WAVEHDR) + rec_bsize);
	if(!rec_htr[rec_count])break;

	rec_htr[rec_count]->dwFlags		    = WHDR_DONE;
	rec_htr[rec_count]->lpData		    = (LPSTR) rec_htr[rec_count]+ sizeof(WAVEHDR);
	rec_htr[rec_count]->dwBufferLength  = rec_bsize;
	rec_htr[rec_count]->dwUser			= rec_count;
    rec_htr[rec_count]->dwBytesRecorded = 0;
    rec_htr[rec_count]->dwLoops			= 1;
    rec_htr[rec_count]->lpNext			= 0;
	rec_ptr[rec_count]=(char*)(rec_htr[rec_count])+sizeof(WAVEHDR);

	if(!waveInPrepareHeader(rec_id, rec_htr[rec_count],sizeof(WAVEHDR)))continue;

	GlobalFreePtr((LPSTR) rec_htr[rec_count]);
	break;
    }

if(rec_count<REC_BUFFER_MIN)				/* memmory error ? */
	{
	RecordClose();
	return 2;
	}


for(x=0;x<rec_count;x++)					/* add buffers */
	{
	waveInAddBuffer( rec_id, rec_htr[x] , sizeof(WAVEHDR) );
	}

rec_close=0;

return 0;
}

/*****************************************************************************
/*
/*		RecordClose
/*
/*****************************************************************************/
/*	Close an audio input
/*	Returns 0 in no error was detected		*/
int RecordClose()
{
int	 err=0;



if(!rec_id)return 1;

rec_close=1;
RecordStop();							/* reset audio input */
waveInReset(rec_id);

if(rec_mixer)mixerClose (rec_mixer);
if(rec_event)CloseHandle(rec_event);

rec_mixer=0;
rec_event=0;
rec_write=0;
rec_read =0;


memset(&rec_wm,0,sizeof(rec_wm));
DeleteCriticalSection(&rec_cs);

while(rec_count>0)						/* release buffers */
	{
	rec_count--;
	if(waveInUnprepareHeader(rec_id ,rec_htr[rec_count],sizeof(WAVEHDR)))err=1;
	if(GlobalFreePtr( (LPSTR) rec_htr[rec_count]))err=1;
	rec_htr[rec_count]=0;
	rec_ptr[rec_count]=0;
	}

if( waveInClose(rec_id   ))err=1;		/* close device */
if(!CloseHandle(rec_event))err=1;;


rec_id=0;
rec_write=0;
rec_read =0;
rec_wait =0;
rec_line_count=-1;


return err;
}


/*****************************************************************************
/*
/*		RecordStop
/*
/*****************************************************************************
/*	Stops recording
/*	Returns 0 in no error was detected		*/
int RecordStop()
{
MMRESULT result;



	rec_on=0;
if(!rec_id)return 1;

result=waveInStop(rec_id);

while(rec_proc)Sleep(50);


if(rec_thread)						/* Close thread */
	{
	int x;
	for(x=0;!rec_th_ex && x<10;x++)Sleep(100);
	if(!rec_th_ex){TerminateThread(rec_thread,0);Sleep(100);}
	CloseHandle(rec_thread);
	rec_thread=0;
	}



return (result==MMSYSERR_NOERROR)? 0:1;
}

/*****************************************************************************
/*
/*		RecordNextBuffer
/*
/*****************************************************************************
/*	Prepare the next buffer for recording */
static void RecordNextBuffer()
{


EnterCriticalSection(&rec_cs);

waveInUnprepareHeader(rec_id, rec_htr[rec_write],sizeof(WAVEHDR));
waveInPrepareHeader  (rec_id, rec_htr[rec_write],sizeof(WAVEHDR));
waveInAddBuffer      (rec_id, rec_htr[rec_write],sizeof(WAVEHDR));


   rec_write++;
if(rec_write>=rec_count)rec_write=0;


LeaveCriticalSection(&rec_cs);
}

/*****************************************************************************
/*
/*		RecordReadIo
/*
/*****************************************************************************
/*	Writes the recording data to the file which is open with Handle
/*	RecordRead uses the write(Handle,void*,int)  function to store the Data
/*	Returns the readed bytes or -1 if an error occurs */
int RecordReadIo(int Handle)
{
return RecordReadProc((int(_cdecl*)(void*,void*,int))write,(void*)Handle);
}

/*****************************************************************************
/*
/*		RecordReadFile
/*
/*****************************************************************************
/*	Writes the recording data to the file which is open with File
/*	RecordRead uses the RecordWriteFile(File,void*,int)  function to store the Data
/*	Returns the readed bytes or -1 if an error occurs */
int RecordReadFile(FILE *File)
{
return RecordReadProc(RecordFileWrite,File);
}
/*****************************************************************************
/*
/*		RecordReadProc
/*
/*****************************************************************************/
/*	Writes the recording data to the funktion Proc
/*	RecordRead uses the Proc(Param,void* Buffer,int Count) to store the Data
/*	Returns the readed bytes or -1 if an error occurs */
int RecordReadProc(RecWrProc Proc,void *Param)
{
int		delta,count,x,y;


if(!rec_id)return -1;


rec_on=1;
rec_proc=1;
waveInStart(rec_id);
count=0;


while(rec_on)								/* while no break */
	{
	EnterCriticalSection(&rec_cs);

	   delta= rec_read - rec_write;
	if(delta< 0)delta += rec_count;
	if(delta>=1 && rec_wait)SetEvent(rec_event);

	LeaveCriticalSection(&rec_cs);


	if(delta<=0)							/* no data ready */
		{
		rec_wait=1;
		WaitForSingleObject(rec_event,4000);
		rec_wait=0;
		continue;
		}

	while(delta>0)							/* write buffer */
		{
		x=rec_htr[rec_write]->dwBytesRecorded;
		y=Proc(Param,rec_htr[rec_write]->lpData,x);
		if(y>0)count+=y;
		RecordNextBuffer();
		delta--;
		}
	}

rec_proc=0;

return count;
}



/*****************************************************************************
/*
/*		RecordReadIoThread
/*
/*****************************************************************************/
/*	This funktion starts an thread which store the Data in a the file
/*	which is opened with Handle
/*	RecordReadThread uses the write(Handle,void*,int) function to store the Data
/*	The Data will be stored in Background */
int RecordReadThreadIo(int Handle)
{

if(!rec_id)return -1;

if(rec_thread)RecordStop();					/* Close old thread */
rrpt_proc  = (int(_cdecl*)(void*,void*,int))write;
rec_thread = CreateThread(0,1024,RecordProcThread,(void*)Handle,0,0);

return 0;
}

/*****************************************************************************
/*
/*		RecordReadFileThread
/*
/*****************************************************************************/
/*	This funktion starts an thread which store the Data in a file
/*	which is opened with File
/*	RecordReadProcThread uses the Proc(Param,void* Buffer,int Count) to store the Data
/*	The Data will be stored in Background */
int RecordReadFileThread(FILE *File)
{
DWORD	id;

if(!rec_id)return -1;

if(rec_thread)RecordStop();					/* Close old thread */

rrpt_proc  = RecordFileWrite;
rec_thread = CreateThread(0,1024,RecordProcThread,File,0,&id);


return 0;
}

/*****************************************************************************
/*
/*		RecordReadProcThread
/*
/*****************************************************************************/
/*	This funktion starts an thread which store the Data with a funktion
/*	RecordReadProcThread uses the Proc(Param,void* Buffer,int Count) to store the Data
/*	The Data will be stored in Background */
int RecordReadProcThread(RecWrProc Proc,void *Param)
{
DWORD	id;

if(!rec_id)return -1;


if(rec_thread)RecordStop();					/* Close old thread */

rrpt_proc  = Proc;
rec_thread = CreateThread(0,1024,RecordProcThread,Param,0,&id);


return 0;
}

/*****************************************************************************
/*
/*		RecordScanInputs
/*
/*****************************************************************************
/*	Scanns the mixer device for input channels
/*	Returns the number of channels or -1 if an error ocures */
int RecordScanInputs()
{
MMRESULT		result;
MIXERLINE		mixer_line;
UINT			i;

	if(!rec_id)return -1;

	if( rec_line_count>=0)return rec_line_count;


	if(!rec_mixer)							/* Open Mixer */
		{
		   result  = mixerGetID( (HMIXEROBJ)rec_id, &rec_mixer_id_en, MIXER_OBJECTF_HWAVEIN );
		if(result != MMSYSERR_NOERROR )	return -1;
		   result  = mixerOpen ( &rec_mixer, rec_mixer_id_en, 0L, 0L, 0L );
		if(result != MMSYSERR_NOERROR )
			{
			mixerClose(rec_mixer);
			rec_mixer=0;
			return -1;
			}


		{
		MIXERCONTROL		Control;
		MIXERLINE			Line;
		MIXERLINECONTROLS	LineControls;
		memset( &Control, 0, sizeof(MIXERCONTROL) );
		Control.cbStruct = sizeof(MIXERCONTROL);

		memset( &LineControls, 0, sizeof(MIXERLINECONTROLS) );
		LineControls.cbStruct = sizeof(MIXERLINECONTROLS);

		memset( &Line, 0, sizeof(MIXERLINE) );
		Line.cbStruct = sizeof(MIXERLINE);

		LineControls.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
		LineControls.dwLineID  = rec_mixer_id_en;
		LineControls.cControls = 1;
		LineControls.cbmxctrl  = sizeof(MIXERCONTROL);
		LineControls.pamxctrl  = &Control;
		result  = mixerGetLineControls( (HMIXEROBJ)rec_mixer, &LineControls, MIXER_GETLINECONTROLSF_ONEBYTYPE );

		rec_mixer_id_vol = Control.dwControlID;
		}


		}


	memset( &mixer_line, 0,sizeof(MIXERLINE) );
	mixer_line.cbStruct  = sizeof(MIXERLINE);
	mixer_line.dwComponentType =  MIXERLINE_COMPONENTTYPE_DST_WAVEIN;


	   result  = mixerGetLineInfo((HMIXEROBJ)rec_mixer,&mixer_line,MIXER_GETLINEINFOF_COMPONENTTYPE );
	if(result != MMSYSERR_NOERROR )
		{
		mixerClose(rec_mixer);
		rec_mixer=0;
		return -1;
		}


	for(i=0;i<mixer_line.cConnections;i++)	/* Scann mixer lines */
		{
		MIXERLINE *line;

		if(i >= REC_MIXERLINE_MAX)break;
		line=rec_line+i;

		memset(line,0,sizeof(MIXERLINE));
		line->cbStruct		= sizeof(MIXERLINE);
		line->dwDestination = mixer_line.dwDestination;
		line->dwSource		= i;

		   result  = mixerGetLineInfo((HMIXEROBJ)rec_mixer,line,MIXER_GETLINEINFOF_SOURCE);
		if(result != MMSYSERR_NOERROR )
			{
			mixerClose( (HMIXER)rec_mixer );
			return -1;	/* FAILURE: Could not get the interated Source line while enumerating.  */
			}

		}


	rec_line_count = i;

return i;
}

/*****************************************************************************
/*
/*		RecordGetInputEnable
/*
/*****************************************************************************
/*	Get the state of a mixer line
/*	If Type >=   0	One of the scanned lines is taken
/*	   Type ==  -1	Serach for MICROPHONE   is the same as REC_MIC
/*	   Type ==  -2	Serach for COMPACTDISC 	is the same as REC_CD
/*	   Type ==  -3	Serach for WAVEOUT 		is the same as REC_WAVEOUT
/*	   Type ==  -4	Serach for ANALOG 		is the same as REC_ANALOG
/*	   Type ==  -5	Serach for AUXILIARY 	is the same as REC_AUX  (Line In)
/*	   Type ==  -6	Serach for DIGITAL 		is the same as REC_DIGITAL
/*	   Type ==  -7	Serach for LINE 		is the same as REC_LINE
/*	   Type ==  -8	Serach for PCSPEAKER 	is the same as REC_PCSPEAKER
/*	   Type ==  -9	Serach for SYNTHESIZER 	is the same as REC_SYNTHESIZER
/*	   Type == -10	Serach for TELEPHONE 	is the same as REC_TELEPHONE
/*	   Type == -11	Serach for UNDEFINED
/*	Returns  0 if the mixerline isn't disabled
/*			 1 if the mixerline is    enabled
/*			-1 if a error was detected	*/
int RecordGetInputEnable(int Type)
{
MIXERCONTROLDETAILS_BOOLEAN		mixer_details_b[REC_MIXERLINE_MAX];
MIXERCONTROLDETAILS				mixer_details;
MMRESULT						result;
DWORD							scan;
INT								i;


	if(!rec_id)return -1;

	RecordScanInputs();						/* Scann all mixer lines */

	if(Type>=0)i=Type;
	else{
		scan=0;

		if(Type ==  -1)scan=MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE;
		if(Type ==  -2)scan=MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
		if(Type ==  -3)scan=MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
		if(Type ==  -4)scan=MIXERLINE_COMPONENTTYPE_SRC_ANALOG;
		if(Type ==  -5)scan=MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY;
		if(Type ==  -6)scan=MIXERLINE_COMPONENTTYPE_SRC_DIGITAL;
		if(Type ==  -7)scan=MIXERLINE_COMPONENTTYPE_SRC_LINE;
		if(Type ==  -8)scan=MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER;
		if(Type ==  -9)scan=MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
		if(Type == -10)scan=MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE;
		if(Type == -11)scan=MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED;

		for(i=0;i<rec_line_count;i++)		/* Serach for line */
			{
			if(rec_line[i].dwComponentType==scan)break;
			}
		}

	if(i>=rec_line_count)return -1;


	mixer_details.cbStruct		 = sizeof(MIXERCONTROLDETAILS);
	mixer_details.dwControlID	 = rec_mixer_id_en;
	mixer_details.cChannels		 = 1;
	mixer_details.cMultipleItems = rec_line_count;
	mixer_details.cbDetails		 = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
	mixer_details.paDetails		 = mixer_details_b;

											/* Get detail info */
	result = mixerGetControlDetails((HMIXEROBJ) rec_mixer,
									&mixer_details,
									MIXER_OBJECTF_HMIXER | MIXER_GETCONTROLDETAILSF_VALUE);

	if(result != MMSYSERR_NOERROR)return -1;


return (mixer_details_b[i].fValue)? 1:0;
}

/*****************************************************************************
/*
/*		RecordSetInputEnable
/*
/*****************************************************************************
/*	Set the mixer line which is active
/*	From this line will be recorded
/*	If Type >=   0	One of the scanned lines is taken
/*	   Type ==  -1	Serach for MICROPHONE   is the same as REC_MIC
/*	   Type ==  -2	Serach for COMPACTDISC 	is the same as REC_CD
/*	   Type ==  -3	Serach for WAVEOUT 		is the same as REC_WAVEOUT
/*	   Type ==  -4	Serach for ANALOG 		is the same as REC_ANALOG
/*	   Type ==  -5	Serach for AUXILIARY 	is the same as REC_AUX  (Line In)
/*	   Type ==  -6	Serach for DIGITAL 		is the same as REC_DIGITAL
/*	   Type ==  -7	Serach for LINE 		is the same as REC_LINE
/*	   Type ==  -8	Serach for PCSPEAKER 	is the same as REC_PCSPEAKER
/*	   Type ==  -9	Serach for SYNTHESIZER 	is the same as REC_SYNTHESIZER
/*	   Type == -10	Serach for TELEPHONE 	is the same as REC_TELEPHONE
/*	   Type == -11	Serach for UNDEFINED
/*	Returns  0 if no error was detected
/*			-1 if a  error was detected	*/
int RecordSetInputEnable(int Type)
{
MIXERCONTROLDETAILS_BOOLEAN		mixer_details_b[REC_MIXERLINE_MAX];
MIXERCONTROLDETAILS				mixer_details;
MMRESULT						result;
DWORD							scan;
INT								i;


	if(!rec_id)return -1;

	RecordScanInputs();						/* Scann all mixer lines */

	if(Type>=0)i=Type;
	else{
		scan=0;

		if(Type ==  -1)scan=MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE;
		if(Type ==  -2)scan=MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
		if(Type ==  -3)scan=MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
		if(Type ==  -4)scan=MIXERLINE_COMPONENTTYPE_SRC_ANALOG;
		if(Type ==  -5)scan=MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY;
		if(Type ==  -6)scan=MIXERLINE_COMPONENTTYPE_SRC_DIGITAL;
		if(Type ==  -7)scan=MIXERLINE_COMPONENTTYPE_SRC_LINE;
		if(Type ==  -8)scan=MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER;
		if(Type ==  -9)scan=MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
		if(Type == -10)scan=MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE;
		if(Type == -11)scan=MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED;

		for(i=0;i<rec_line_count;i++)		/* Serach for line */
			{
			if(rec_line[i].dwComponentType==scan)break;
			}
		}

	if(i>=rec_line_count)return -1;

	memset(mixer_details_b,0,sizeof(mixer_details_b));
	mixer_details.cbStruct		 = sizeof(MIXERCONTROLDETAILS);
	mixer_details.dwControlID	 = rec_mixer_id_en;
	mixer_details.cChannels		 = 1;
	mixer_details.cMultipleItems = rec_line_count;
	mixer_details.cbDetails		 = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
	mixer_details.paDetails		 = mixer_details_b;
	mixer_details_b[i].fValue    = 1;


	result = mixerSetControlDetails((HMIXEROBJ) rec_mixer,
									&mixer_details,
									MIXER_OBJECTF_HMIXER | MIXER_SETCONTROLDETAILSF_VALUE);

	if(result == MMSYSERR_NOERROR)return 0;


return -1;
}

/*****************************************************************************
/*
/*		RecordGetInputVolume
/*
/*****************************************************************************
/*	Get the recording volume level of a mixer line
/*	If Type >=   0	One of the scanned lines is taken
/*	   Type ==  -1	Serach for MICROPHONE   is the same as REC_MIC
/*	   Type ==  -2	Serach for COMPACTDISC 	is the same as REC_CD
/*	   Type ==  -3	Serach for WAVEOUT 		is the same as REC_WAVEOUT
/*	   Type ==  -4	Serach for ANALOG 		is the same as REC_ANALOG
/*	   Type ==  -5	Serach for AUXILIARY 	is the same as REC_AUX  (Line In)
/*	   Type ==  -6	Serach for DIGITAL 		is the same as REC_DIGITAL
/*	   Type ==  -7	Serach for LINE 		is the same as REC_LINE
/*	   Type ==  -8	Serach for PCSPEAKER 	is the same as REC_PCSPEAKER
/*	   Type ==  -9	Serach for SYNTHESIZER 	is the same as REC_SYNTHESIZER
/*	   Type == -10	Serach for TELEPHONE 	is the same as REC_TELEPHONE
/*	   Type == -11	Serach for UNDEFINED
/*	Returns  Volume=0 to 0xFFFF
/*			-1 if a error was detected	*/
int RecordGetInputVolume(int Type)
{
MIXERCONTROLDETAILS_UNSIGNED mxcdVolume;
MIXERCONTROLDETAILS			 mxcd;
MIXERLINECONTROLS			 mxlc;
MIXERCONTROL				 mxc;
MMRESULT					 result;
DWORD						 scan;
INT							 i;



	if(!rec_id)return -1;

	RecordScanInputs();						/* Scann all mixer lines */

	if(Type>=0)i=Type;
	else{
		scan=0;

		if(Type ==  -1)scan=MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE;
		if(Type ==  -2)scan=MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
		if(Type ==  -3)scan=MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
		if(Type ==  -4)scan=MIXERLINE_COMPONENTTYPE_SRC_ANALOG;
		if(Type ==  -5)scan=MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY;
		if(Type ==  -6)scan=MIXERLINE_COMPONENTTYPE_SRC_DIGITAL;
		if(Type ==  -7)scan=MIXERLINE_COMPONENTTYPE_SRC_LINE;
		if(Type ==  -8)scan=MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER;
		if(Type ==  -9)scan=MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
		if(Type == -10)scan=MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE;
		if(Type == -11)scan=MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED;

		for(i=0;i<rec_line_count;i++)		/* Serach for line */
			{
			if(rec_line[i].dwComponentType==scan)break;
			}
		}

	if(i>=rec_line_count)return -1;



	/* get dwControlID*/
	mxlc.cbStruct = sizeof(MIXERLINECONTROLS);
	mxlc.dwLineID = rec_line[i].dwLineID;
	mxlc.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
	mxlc.cControls = 1;
	mxlc.cbmxctrl = sizeof(MIXERCONTROL);
	mxlc.pamxctrl = &mxc;

	   result  = mixerGetLineControls((HMIXEROBJ)rec_mixer,&mxlc,MIXER_OBJECTF_HMIXER|MIXER_GETLINECONTROLSF_ONEBYTYPE);
	if(result != MMSYSERR_NOERROR){return -1;}



	mxcd.cbStruct		= sizeof(MIXERCONTROLDETAILS);
	mxcd.dwControlID	= mxc.dwControlID;
	mxcd.cChannels		= 1;
	mxcd.cMultipleItems = 0;
	mxcd.cbDetails		= sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	mxcd.paDetails		= &mxcdVolume;


	   result  = mixerGetControlDetails((HMIXEROBJ)rec_mixer,&mxcd, MIXER_OBJECTF_HMIXER|MIXER_GETCONTROLDETAILSF_VALUE);
	if(result != MMSYSERR_NOERROR){return -1;}


return mxcdVolume.dwValue;
}


/*****************************************************************************
/*
/*		RecordSetInputVolume
/*
/*****************************************************************************
/*	Set the recording volume level of a mixer line
/*	If Type >=   0	One of the scanned lines is taken
/*	   Type ==  -1	Serach for MICROPHONE   is the same as REC_MIC
/*	   Type ==  -2	Serach for COMPACTDISC 	is the same as REC_CD
/*	   Type ==  -3	Serach for WAVEOUT 		is the same as REC_WAVEOUT
/*	   Type ==  -4	Serach for ANALOG 		is the same as REC_ANALOG
/*	   Type ==  -5	Serach for AUXILIARY 	is the same as REC_AUX  (Line In)
/*	   Type ==  -6	Serach for DIGITAL 		is the same as REC_DIGITAL
/*	   Type ==  -7	Serach for LINE 		is the same as REC_LINE
/*	   Type ==  -8	Serach for PCSPEAKER 	is the same as REC_PCSPEAKER
/*	   Type ==  -9	Serach for SYNTHESIZER 	is the same as REC_SYNTHESIZER
/*	   Type == -10	Serach for TELEPHONE 	is the same as REC_TELEPHONE
/*	   Type == -11	Serach for UNDEFINED
/*	Returns  Volume=0 to 0xFFFF
/*			-1 if a error was detected	*/
int RecordSetInputVolume(int Type,unsigned Volume)
{
MIXERCONTROLDETAILS_UNSIGNED mxcdVolume;
MIXERCONTROLDETAILS			 mxcd;
MIXERLINECONTROLS			 mxlc;
MIXERCONTROL				 mxc;
MMRESULT					 result;
DWORD						 scan;
INT							 i;



	if(!rec_id)return -1;

	RecordScanInputs();						/* Scann all mixer lines */

	if(Type>=0)i=Type;
	else{
		scan=0;

		if(Type ==  -1)scan=MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE;
		if(Type ==  -2)scan=MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
		if(Type ==  -3)scan=MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
		if(Type ==  -4)scan=MIXERLINE_COMPONENTTYPE_SRC_ANALOG;
		if(Type ==  -5)scan=MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY;
		if(Type ==  -6)scan=MIXERLINE_COMPONENTTYPE_SRC_DIGITAL;
		if(Type ==  -7)scan=MIXERLINE_COMPONENTTYPE_SRC_LINE;
		if(Type ==  -8)scan=MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER;
		if(Type ==  -9)scan=MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
		if(Type == -10)scan=MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE;
		if(Type == -11)scan=MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED;

		for(i=0;i<rec_line_count;i++)		/* Serach for line */
			{
			if(rec_line[i].dwComponentType==scan)break;
			}
		}

	if(i>=rec_line_count)return -1;



	/* get dwControlID*/
	mxlc.cbStruct = sizeof(MIXERLINECONTROLS);
	mxlc.dwLineID = rec_line[i].dwLineID;
	mxlc.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
	mxlc.cControls = 1;
	mxlc.cbmxctrl = sizeof(MIXERCONTROL);
	mxlc.pamxctrl = &mxc;

	   result  = mixerGetLineControls((HMIXEROBJ)rec_mixer,&mxlc,MIXER_OBJECTF_HMIXER|MIXER_GETLINECONTROLSF_ONEBYTYPE);
	if(result != MMSYSERR_NOERROR)return -1;


	mxcdVolume.dwValue	= Volume;
	mxcd.cbStruct		= sizeof(MIXERCONTROLDETAILS);
	mxcd.dwControlID	= mxc.dwControlID;
	mxcd.cChannels		= 1;
	mxcd.cMultipleItems = 0;
	mxcd.cbDetails		= sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	mxcd.paDetails		= &mxcdVolume;


	   result  = mixerSetControlDetails((HMIXEROBJ)rec_mixer,&mxcd, MIXER_OBJECTF_HMIXER|MIXER_GETCONTROLDETAILSF_VALUE);
	if(result != MMSYSERR_NOERROR)return -1;



return mxcdVolume.dwValue;
}


/*****************************************************************************
/*
/*		RecordGetInputName
/*
/*****************************************************************************/
/*	Gets the name of a input
/*	If Type >=   0	One of the scanned lines is taken
/*	   Type ==  -1	Serach for MICROPHONE   is the same as REC_MIC
/*	   Type ==  -2	Serach for COMPACTDISC 	is the same as REC_CD
/*	   Type ==  -3	Serach for WAVEOUT 		is the same as REC_WAVEOUT
/*	   Type ==  -4	Serach for ANALOG 		is the same as REC_ANALOG
/*	   Type ==  -5	Serach for AUXILIARY 	is the same as REC_AUX  (Line In)
/*	   Type ==  -6	Serach for DIGITAL 		is the same as REC_DIGITAL
/*	   Type ==  -7	Serach for LINE 		is the same as REC_LINE
/*	   Type ==  -8	Serach for PCSPEAKER 	is the same as REC_PCSPEAKER
/*	   Type ==  -9	Serach for SYNTHESIZER 	is the same as REC_SYNTHESIZER
/*	   Type == -10	Serach for TELEPHONE 	is the same as REC_TELEPHONE
/*	   Type == -11	Serach for UNDEFINED
/*	Returns  the name or 0 if no input was found */
char *RecordGetInputName  (int Type,int ShortName)
{
DWORD						 scan;
INT							 i;



	if(!rec_id)return 0;

	RecordScanInputs();						/* Scann all mixer lines */

	if(Type>=0)i=Type;
	else{
		scan=0;

		if(Type ==  -1)scan=MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE;
		if(Type ==  -2)scan=MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
		if(Type ==  -3)scan=MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
		if(Type ==  -4)scan=MIXERLINE_COMPONENTTYPE_SRC_ANALOG;
		if(Type ==  -5)scan=MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY;
		if(Type ==  -6)scan=MIXERLINE_COMPONENTTYPE_SRC_DIGITAL;
		if(Type ==  -7)scan=MIXERLINE_COMPONENTTYPE_SRC_LINE;
		if(Type ==  -8)scan=MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER;
		if(Type ==  -9)scan=MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
		if(Type == -10)scan=MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE;
		if(Type == -11)scan=MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED;

		for(i=0;i<rec_line_count;i++)		/* Serach for line */
			{
			if(rec_line[i].dwComponentType==scan)break;
			}
		}

	if(i>=rec_line_count)return 0;

return (ShortName)? rec_line[i].szShortName:rec_line[i].szName;
}

/*****************************************************************************
/*
/*		RecordInfo
/*
/*****************************************************************************/
/*	Get infos of the opened record input
/*	Freq		: Target for the sampling frequency (8000,11025,22050,44100)
/*	Bits		: Target for the bit-count for one value (8,12,16)
/*	Channels	: Target for the number of input-channels (1=mono,2=stereo)
/*	BufferSize	: Target for the maximum buffer size for one write operation
/*  If a pointer is zero then the there will not saved a value.
/*	Return	0	: Driver is not open
/*			1	: Driver is open				 */
int	RecordInfo(unsigned *Freq,unsigned *Bits,unsigned *Channels,unsigned  *BufferSize )
{

if(Freq      )*Freq		  = rec_wm.nSamplesPerSec;
if(Bits      )*Bits		  = rec_wm.wBitsPerSample;
if(Channels	 )*Channels   = rec_wm.nChannels;
if(BufferSize)*BufferSize = rec_bsize;

return (rec_mixer)? 1:0;
}


/*****************************************************************************
/*
/*		RecordIsOpen
/*
/*****************************************************************************/
/*	Returns 1 ist the Record Driver ist open                                 */
int RecordIsOpen()
{
return (rec_mixer)? 1:0;
}
